#! /usr/bin/env python
"""
Copyright (c) 2019-2020 Jolle Jolles <j.w.jolles@gmail.com>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at:

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function

import os
import cv2
import ast
import sys
import time
import glob
import argparse
import subprocess

from threading import Thread
from multiprocess import Pool
from pythutils.sysutils import lineprint
from pythutils.drawutils import draw_text
from pythutils.fileutils import listfiles, get_ext, commonpref, move
from pythutils.mediautils import get_vid_params, videowriter, imgresize

class KeyboardInterruptError(Exception): pass

class Convert:

    """
    CASDonverter class to convert a directory of videos to mp4 with
    potential to write frame number on each frames

    Parameters
    -----------
    indir : str, default = ""
        Directory containing the videos
    outdir : str, default = ""
        Directory where the converted videos should be stored
    type : str, default = ".h264"
        The filetype of the media to convert
    withframe : bool, default = False
        Type of conversion, either very fast conversion using ffmpeg or
        using opencv to add frame number to each frame
    delete : bool, default = False
        If the original videos should be delete or not
    pools : int, default = 4
        Number of computer cores to use for conversion script
    resizeval : float, default = 1
        Float value to which video should be resized
    imgfps : int, default = 25
        Framerate for conversion of images to video
    sleeptime : 2, default = None
        Time in seconds between subsequent checks of file folder. No only check
        once use default
    """

    def __init__(self, indir="", outdir="", type=".h264", withframe=False,
                 overwrite=False, delete=False, pools=4, resizeval=1, fps=None,
                 imgfps=25, internal=False, sleeptime=None, interrupt=True):

        if internal:
            lineprint("Running convert function..")

        global interrupted
        interrupted = False
        if interrupt:
            def keythread():
                global interrupted
                input()
                interrupted = True
            Thread(target=keythread).start()

        self.indir = os.getcwd() if indir == "" else indir
        assert os.path.exists(self.indir), "in-directory does not exist.."
        self.outdir = self.indir if outdir == "" else outdir
        assert os.path.exists(self.outdir), "out-directory does not exist.."

        self.rootdir = os.getcwd()
        os.chdir(self.indir)

        self.type = type
        self.withframe = withframe
        self.delete = delete
        self.pools = int(pools)
        self.resizeval = float(resizeval)
        self.fps = int(fps) if fps is not None else None
        self.imgfps = int(imgfps)

        while not interrupted:
            files = listfiles(self.indir, self.type, keepdir = False)
            old = listfiles(self.indir, self.type, keepext = False)
            new = listfiles(self.outdir, ".mp4", keepext = False)
            self.todo = files
            if not overwrite:
                self.todo = [files[i] for i,file in enumerate(old) if file not in new]
            if self.type in [".jpg",".jpeg",".png"] and len(self.todo)>0:
                 if len([f for f in new if commonpref(self.todo) in f])>0:
                     self.todo = []
            self.convertpool()
            msg = "No files to convert.."
            if sleeptime == None:
                lineprint(msg)
                break
            else:
                lineprint(msg+" rechecking in "+str(sleeptime)+"s..")
                time.sleep(sleeptime)


    def conv_single(self, filein):

        try:
            filebase = os.path.basename(filein)
            fileout = filein if self.outdir == "" else self.outdir+"/"+filebase
            lineprint("Start converting "+filebase)

            if self.withframe:
                vid = cv2.VideoCapture(filein)
                fps, width, height, _ = get_vid_params(vid)
                if self.fps is None:
                    self.fps = fps
                print(self.fps)
                vidout = videowriter(fileout, width, height, self.fps, self.resizeval)

                while True:
                    flag, frame = vid.read()
                    if flag:
                        if self.resizeval != 1:
                            frame = imgresize(frame, self.resizeval)
                        frame_nr = int(vid.get(cv2.CAP_PROP_POS_FRAMES))
                        draw_text(frame, str(frame_nr), (10,10), 0.9)
                        vidout.write(frame)
                    if not flag:
                        break

            else:
                if self.resizeval != 1:
                    comm = "' -vf 'scale=iw*" + str(self.resizeval) + ":-2' '"
                else:
                    comm = "' -vcodec copy '"
                bashcomm = "ffmpeg"
                if self.fps is not None:
                    bashcomm = bashcomm+" -r "+ str(self.fps)
                bashcomm = bashcomm+" -i '"+filein+comm+fileout[:-len(self.type)]+".mp4'"
                bashcomm = bashcomm + " -y -nostats -loglevel 0"
                output = subprocess.check_output(['bash','-c', bashcomm])

            lineprint("Finished converting "+filebase)

        except KeyboardInterrupt:
            raise KeyboardInterruptError()


    def convertpool(self):

        if len(self.todo) > 0:

            if self.type in [".h264",".mp4",".avi"]:

                pool = Pool(min(self.pools, len(self.todo)))
                try:
                    pool.map(self.conv_single, self.todo)
                    pool.close()
                    lineprint("Done converting all videofiles!")
                except KeyboardInterrupt:
                    lineprint("User terminated converting pool..")
                    pool.terminate()
                except Exception as e:
                    excep = "Got exception: %r, terminating pool" % (e,)
                    lineprint(excep)
                    pool.terminate()
                finally:
                    pool.join()

                if self.delete:
                    for filein in self.todo:
                        os.remove(filein)
                    lineprint("Deleted all original videofiles..")

            elif self.type in [".jpg",".jpeg",".png"]:

                vidname = commonpref(self.todo)
                lineprint("Start converting "+str(len(self.todo))+" images")

                frame_array = []
                for filename in self.todo:
                    frame = cv2.imread(filename)
                    frame_array.append(frame)
                    #os.rename(filename, self.outdir+"/"+filename)
                h, w, _ = frame_array[0].shape
                if self.outdir != "":
                    vidname = self.outdir+"/"+os.path.basename(vidname)
                vidout = videowriter(vidname, w, h, self.imgfps, self.resizeval)
                for i in range(len(frame_array)):
                    vidout.write(frame_array[i])
                vidout.release()
                lineprint("Finished converting "+os.path.basename(vidname))

            else:
                lineprint("No video or image files found..")


def conv():

    """To run the convert function from the command line"""

    parser = argparse.ArgumentParser(prog="Convert", description=Convert.__doc__,
             formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument("-i", "--indir" ,default="", metavar="")
    parser.add_argument("-o", "--outdir" ,default="", metavar="")
    parser.add_argument("-t", "--type", default=".h264", metavar="")
    parser.add_argument("-w", "--withframe", default="False", metavar="")
    parser.add_argument("-d", "--delete", default="False", metavar="")
    parser.add_argument("-p", "--pools", default=4, metavar="")
    parser.add_argument("-r", "--resizeval", default=1, metavar="")
    parser.add_argument("-f", "--imgfps", default=25, metavar="")

    args = parser.parse_args()
    args.withframe = ast.literal_eval(args.withframe)
    args.delete = ast.literal_eval(args.delete)
    Convert(indir=args.indir, outdir=args.outdir, type=args.type,
            withframe=args.withframe, delete=args.delete, pools=args.pools,
            resizeval=args.resizeval, imgfps=args.imgfps)



#! /usr/bin/env python
"""
Copyright (c) 2019 - 2020 Jolle Jolles <j.w.jolles@gmail.com>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at:

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This function was partly based on code provided by Dave Jones in a
reply on a question posted on stackoverflow: https://bit.ly/2V49f48
"""

from __future__ import print_function

import cv2
import numpy as np

def setgains(attempts = 100, step = 0.05, startgains = (1,2),
             zoom = (0,0,1,1), auto = True):

    """
    Find the best gains for the raspberry pi camera

    Parameters
    ----------
    attempts : int, default = 100
        Number of times gains should be shifted when in automatic mode
    step : float, detault = 0.05
        The stepsize between the different gains
    startgains : tuple, default = (1,2)
        The starting gains
    zoom : tuple, default = (0,0,1,1)
        The camera zoom (x,y,w,h)
    auto : bool, default = True
        If gains should be determined automatically or manually

    Returns
    -------
    (rg,bg) : A tuple of red and blue camera gains
    """

    #load picamera module in-function so pirecorder is installable on all OS
    import picamera
    import picamera.array

    camera = picamera.PiCamera()
    height, width = (480, 640)
    camera.resolution = (width, height)
    camera.zoom = zoom
    camera.awb_mode = 'off'
    rg, bg = startgains
    camera.awb_gains = (rg, bg)

    if auto:
        with picamera.array.PiRGBArray(camera, size=(128, 72)) as output:
            for i in range(attempts):
                camera.capture(output, format="rgb", resize=(128, 80), use_video_port=True)
                r, g, b = (np.mean(output.array[..., i]) for i in range(3))
                print("R:%5.2f, B:%5.2f = (%5.2f, %5.2f, %5.2f)" % (rg, bg, r, g, b))
                if abs(r - g) > 2:
                    if r > g:
                        rg -= step
                    else:
                        rg += step
                if abs(b - g) > 1:
                    if b > g:
                        bg -= step
                    else:
                        bg += step
                camera.awb_gains = (rg, bg)
                output.seek(0)
                output.truncate()
    else:
        cv2.namedWindow("Image")
        while True:
            image = np.empty((height * width * 3,), dtype=np.uint8)
            camera.capture(image, "bgr")
            image = image.reshape((height, width, 3))
            cv2.imshow("Image", image)
            print("R:%5.2f, B:%5.2f" % (rg, bg))
            camera.awb_gains = (rg, bg)

            k = cv2.waitKey(1) & 0xFF
            if k == ord("q"):
                rg = round(max(rg-step,0.5),2)
            if k == ord("w"):
                rg = round(min(rg+step,2.5),2)
            if k == ord("e"):
                bg = round(max(bg-step,0.5),2)
            if k == ord("r"):
                bg = round(min(bg+step,2.5),2)
            if k == ord("s"):
                print("User exited..")
                break
            if k == 27:
                rg, bg = startgains
                print("User escaped..")
                break

        cv2.waitKey(1)
        cv2.destroyAllWindows()

    camera.close()

    return (rg, bg)


if __name__ == "__main__":
      setgains()



#! /usr/bin/env python
"""
Copyright (c) 2019 - 2020 Jolle Jolles <j.w.jolles@gmail.com>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at:

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import cv2

from .videoin import VideoIn
from pythutils.mediautils import add_transimg, imgresize

def stream(rotation = 0):

    """
    Opens a video stream of the rpi camera. With keypress 'f' the user can make
    the videostream fullscreen and with the 'ESC' key the window is closed.
    """

    fullscreen = False
    vid = VideoIn(vidsize=0.25, rotation=rotation).start()
    cv2.namedWindow("Stream", cv2.WND_PROP_FULLSCREEN)
    while True:
        frame = vid.read()
        cv2.imshow('Stream', frame)
        k = cv2.waitKey(1) & 0xFF
        if k == ord("f"):
            fullscreen = not fullscreen
            if fullscreen:
                cv2.setWindowProperty("Stream", cv2.WND_PROP_FULLSCREEN,
                                      cv2.WINDOW_FULLSCREEN)
            else:
                cv2.setWindowProperty("Stream",cv2.WND_PROP_AUTOSIZE,
                                               cv2.WINDOW_NORMAL)
                cv2.resizeWindow("Stream", vid.res[0], vid.res[1])

        if k == 27:
            break

    vid.stop()
    cv2.destroyAllWindows()
    cv2.waitKey(1)


def overlay_stream(imagefile = "", alpha = 0.5, rotation = 0):

    # Check if image file loads
    assert os.path.isfile(imagefile), "Image file could not be loaded.."

    # Start videostream
    fullscreen = False
    vid = VideoIn(vidsize=1, rotation=rotation).start()
    frame = vid.read()
    overlay = frame.copy()
    h, w, _ = frame.shape
    cv2.namedWindow("Stream", cv2.WND_PROP_FULLSCREEN)

    # Load and resize the image to fit
    photo = cv2.imread(imagefile)
    photo = imgresize(photo, resize = 1, dims = (w,h))

    # Draw photo on overlay
    overlay[0:h,0:w] = photo

    # Start the loop
    c = 0
    vid.stop()
    vid = VideoIn(vidsize=1, rotation=rotation).start()
    while True:

        c += 1
        print(c)
        # Extract the frame
        frame = vid.read()
        output = frame.copy()

        # Draw overlay semi-transparent on frame
        # cv2.addWeighted(overlay, alpha, output, 1 - alpha, 0, output)

        # Draw in black and white
        output = cv2.cvtColor(output, cv2.COLOR_BGR2GRAY)

        # Show the stream
        cv2.imshow("Stream", output)
        k = cv2.waitKey(1) & 0xFF
        if k == ord("["):
            alpha = max(alpha-0.05, 0)
        if k == ord("]"):
            alpha = min(alpha+0.05, 1)
        if k == ord("f"):
            fullscreen = not fullscreen
            if fullscreen:
                cv2.setWindowProperty("Stream", cv2.WND_PROP_FULLSCREEN,
                                      cv2.WINDOW_FULLSCREEN)
            else:
                cv2.setWindowProperty("Stream",cv2.WND_PROP_AUTOSIZE,
                                               cv2.WINDOW_NORMAL)
                cv2.resizeWindow("Stream", vid.res[0], vid.res[1])

        if k == 27:
            break

    vid.stop()
    cv2.destroyAllWindows()
    cv2.waitKey(1)


if __name__ == "__main__":
      stream()



#! /usr/bin/env python
"""
Copyright (c) 2015 - 2020 Jolle Jolles <j.w.jolles@gmail.com>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at:

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function
from builtins import input

import os
import sys
import yaml

import argparse
import numpy as np
from io import BytesIO
from ast import literal_eval
from datetime import datetime
from socket import gethostname
from fractions import Fraction
from time import sleep, strftime
from localconfig import LocalConfig
from pythutils.sysutils import Logger, lineprint, homedir, checkfrac
from pythutils.fileutils import name
from pythutils.mediautils import picamconv

from .schedule import Schedule
from .setgains import setgains
from .calibrate import Calibrate
from .__version__ import __version__

class PiRecorder:

    """
    Recorder class for setting up the rpi for controlled image & video recording

    Parameters
    ----------
    recdir : str, default = "pirecorder/recordings"
        The directory where media will be stored. Default is "recordings". If
        different, a folder with name corresponding to location will be created
        inside the home directory. If no name is provided (""), the files are
        stored in the home directory. If "NAS" is provided it will additionally
        check if the folder links to a mounted drive.
    label : str, default = "test"
        Label for associating with the recording and stored in the filenames.
    rectype : ["img", "imgseq", "vid", "vidseq"], default = "img"
        Recording type, either a single image or video or a sequence of images
        or videos.

    Config settings
    ---------------
    rotation : int, default = 0
        Custom rotation specific to the Raspberry Pi, should be either 0 or 180.
    brighttune : int, default = 0
        A rpi-specific brightness compensation factor to standardize light levels
        across multiple rpi's, an integer between -10 and 10.
    roi : tuple, default = None
        Region of interest to be used for recording. Consists of coordinates of
        topleft and bottom right coordinate of a rectangular area encompassing
        the region of interest. Can be set with the set_roi() method.
    gains : tuple, default = (1.0, 2.5)
        Custom gains specific to the Raspberry PI to set the colorspace. The
        gains for an ideal white balance can be automatically set with the
        get_gains() method.

    brightness : int, default = 45
        The brightness level of the camera, an integer value between 0 and 100.
    contrast : int, default = 20
        The image contrast, an integer value between 0 and 100.
    saturation : int, default -100
        The color saturation level of the image, an integer value between -100
        and 100.
    iso : int, default = 200
        The camera ISO value, an integer value in sequence [200,400,800,1600].
        Higher values are more light sensitive but have higher gain.
    sharpness : int, default = 50
        The sharpness of the camera, an integer value between -100 and 100.
    compensation : int, default = 0
        Camera lighting compensation. Ranges between 0 and 20. Compensation
        artificially adds extra light to the image.
    shutterspeed : int, detault = 10000
        Shutter speed of the camera in microseconds, i.e. the default of 10000
        is equivalent to 1/100th of a second. A longer shutterspeed will result
        in a brighter image but more motion blur. Important: the framerate of
        the camera will be adjusted based on the shutterspeed. At shutter-
        speeds above ~ 0.2s this results in increasingly longer waiting times
        between images so a standard imgwait time should be chosen that is 6+
        times more than the shutterspeed. For example, for a shutterspeed of
        300000 imgwait should be > 1.8s.
    imgdims : tuple, default = (2592, 1944)
        The resolution of the images to be taken in pixels. The default is the
        max resolution that does not return an error for this mode for the v1.5
        rpi camera. Note that rpi camera v2 has a much higher maximum resolution
        of 3280 x 2464.
    viddims : tuple, default = (1640,1232)
        The resolution of the videos to be taken in pixels. The default is the
        max resolution that does not return an error for this mode.
    imgfps : int, default = 1
        The framerate for recording images. Will be set automatically based on
        the imgwait setting so should not be set by user.
    vidfps : int, default = 24
        The framerate for recording video.
    imgwait : float, default = 5.0
    	The delay between subsequent images in seconds. When a delay is provided
      	that is less than ~x5 the shutterspeed, the camera processing time will
        take more time than the provided imgwait parameter and so images are
        taken immideately one after the other. To take a sequence of images at
        the exact right delay interval the imgwait parameter should be at least
        5x the shutterspeed (e.g. shutterspeed of 400ms needs imgwait of 2s).
    imgnr : int, default = 12
        The number of images that should be taken. When this number is reached,
        the recorder will automatically terminate.
    imgtime : integer, default = 60
        The time in seconds during which images should be taken. The minimum of
        a) imgnr and b) nr of images based on imgwait and imgtime will be
        selected.
    imgquality : int, default = 50
        Specifies the quality that the jpeg encoder should attempt to maintain.
        Use values between 1 and 100, where higher values are higher quality.
    vidduration : int, default = 10
        Duration of video recording in seconds.
    viddelay : int, default = 0
        Extra recording time in seconds that will be added to vidduration. Its
        use is for filming acclimatisation time that can then easily be cropped
        for tracking.
    vidquality : int, default = 11
        Specifies the quality that the h264 encoder should attempt to maintain.
        Use values between 10 and 40, where 10 is extremely high quality, and
        40 is extremely low.

    Output
    -------
    Either one or multiple .h264 or .jpg files. All files are automatically
    named according to the label, the host name, date, time and potentially
    session number or count nr, e.g.
    - single image: 'pilot_180312_PI13_101300.jpg
    - multiple images: 'pilot_180312_PI13_img00231_101300.jpg
    - video: 'pilot_180312_PI13_S01_101300.h264

    Returns
    -------
    self : class
        Recorder class instance
    """

    def __init__(self, configfile = "pirecorder.conf", logging=True):

        lineprint("pirecorder "+__version__+" started!")
        lineprint("="*47, False)

        self.system = "auto"
        self.host = gethostname()
        self.home = homedir()
        self.setupdir = self.home + "pirecorder"
        self.logfolder = self.setupdir+"/logs"
        if not os.path.exists(self.logfolder):
            os.makedirs(self.setupdir)
            os.makedirs(self.logfolder)
            lineprint("Setup folder created ("+self.setupdir+")..")
        if not os.path.exists(self.logfolder):
            lineprint("Setup folder already exists but was not set up properly..")

        if logging:
            self.log = Logger(self.logfolder+"/pirecorder.log")
            self.log.start()

        self.brightfile = self.setupdir+"/cusbright.yml"
        self.configfile = self.setupdir+"/"+configfile

        self.config = LocalConfig(self.configfile, compact_form = True)
        if not os.path.isfile(self.configfile):
            lineprint("Config file not found, new file created..")
            for section in ["rec","cam","cus", "img","vid"]:
                if section not in list(self.config):
                    self.config.add_section(section)
            self.set_config(recdir="pirecorder/recordings", subdirs=False, label="test",
                            rectype="vid", rotation=0, brighttune=0, roi=None,
                            gains=(1.0,2.5), brightness=45, contrast=10,
                            saturation=-100, iso=200, sharpness=0, compensation=0,
                            shutterspeed=8000, imgdims=(2592,1944),
                            viddims=(1640,1232), imgfps=1, vidfps=24, imgwait=5.0,
                            imgnr=12, imgtime=60, imgquality=50, vidduration=10,
                            viddelay=10, vidquality=11, internal="")
            lineprint("Config settings stored..")

        else:
            lineprint("Config file " + configfile + " loaded..")
            lineprint("Recording " + self.config.rec.rectype + " in " +\
                          self.home + self.config.rec.recdir)

        self._imgparams()
        self._shuttertofps()
        if self.config.rec.rectype == "imgseq":
            if self.config.cam.shutterspeed/1000000. <= (self.config.img.imgwait/5):
                lineprint("imgwait is not enough for provided shutterspeed" + \
                          ", will be overwritten..")

        if self.config.rec.recdir == "NAS":
            if not os.path.ismount(self.config.rec.recdir):
                self.recdir = self.home
                lineprint("Recdir not mounted, storing in home directory..")
        self.recdir = self.home + self.config.rec.recdir
        if not os.path.exists(self.recdir):
            os.makedirs(self.recdir)

        os.chdir(self.recdir)


    def _setup_cam(self):

        """Sets up the raspberry pi camera based on configuration"""

        #load picamera module in-function so pirecorder is installable on all OS
        import picamera
        import picamera.array

        self.cam = picamera.PiCamera()
        self.cam.rotation = self.config.cus.rotation
        self.cam.exposure_compensation = self.config.cam.compensation

        if self.config.rec.rectype in ["img","imgseq"]:
            self.cam.resolution = literal_eval(self.config.img.imgdims)
            self.cam.framerate = self.config.img.imgfps
        if self.config.rec.rectype in ["vid","vidseq"]:
            self.cam.resolution = literal_eval(self.config.vid.viddims)
            self.cam.framerate = self.config.vid.vidfps
        self.rawCapture = picamera.array.PiRGBArray(self.cam, size = self.cam.resolution)

        if self.config.cus.roi is None:
            self.cam.zoom = (0,0,1,1)
            self.resize = self.cam.resolution
        else:
            self.cam.zoom = literal_eval(self.config.cus.roi)
            w = int(self.cam.resolution[0]*self.cam.zoom[2])
            h = int(self.cam.resolution[1]*self.cam.zoom[3])
            self.resize = picamconv((w,h))

        self.longexpo = False if self.cam.framerate >= 6 else True
        if self.longexpo:
            lineprint("Long exposure, warming up camera..")
            sleep(6) if self.cam.framerate > 1.6 else sleep(10)
        else:
            lineprint("Camera warming up..")
            sleep(2)

        self.cam.shutter_speed = self.config.cam.shutterspeed
        self.cam.exposure_mode = "off"
        self.cam.awb_mode = "off"
        self.cam.awb_gains = checkfrac(self.config.cus.gains)
        brightness = self.config.cam.brightness + self.config.cus.brighttune
        self.cam.brightness = brightness

        self.cam.contrast = self.config.cam.contrast
        self.cam.saturation = self.config.cam.saturation
        self.cam.iso = self.config.cam.iso
        self.cam.sharpness = self.config.cam.sharpness


    def _imgparams(self, mintime = 0.45):

        """
        Calculates minimum possible imgwait and imgnr based on imgtime. The
        minimum time between subsequent images is by default set to 0.45s, the
        time it takes to take an image with max resolution.
        """

        self.config.img.imgwait = max(mintime, self.config.img.imgwait)
        totimg = int(self.config.img.imgtime / self.config.img.imgwait)
        self.config.img.imgnr = min(self.config.img.imgnr, totimg)


    def _shuttertofps(self, minfps = 1, maxfps = 40):

        """Computes image fps based on shutterspeed within provided range"""

        fps = 1./(self.config.cam.shutterspeed/1000000.)
        fps = max(fps, minfps)
        self.config.img.imgfps = min(fps, maxfps)


    def _namefile(self):

        """
        Provides a filename for the media recorded. Filenames include label,
        date, rpi name, and time. Images part of image sequence additionally
        contain a sequence number. e.g. test_180708_pi12_S01_100410
        """

        imgtypes = ["img","imgseq"]
        self.filetype = ".jpg" if self.config.rec.rectype in imgtypes else ".h264"

        if self.config.rec.rectype == "imgseq":
            date = strftime("%y%m%d")
            counter = "im{counter:05d}" if self.config.img.imgnr>999 else "im{counter:03d}"
            time = "{timestamp:%H%M%S}"
            self.filename = "_".join([self.config.rec.label,date,self.host,counter,time])
            self.filename = self.filename+self.filetype
        else:
            date = strftime("%y%m%d")
            self.filename = "_".join([self.config.rec.label, date, self.host])+"_"

        if self.config.rec.subdirs:
            subdir = name("_".join([self.config.rec.label,date,self.host]))
            os.makedirs(subdir, exist_ok=True)
            self.filename = subdir+"/"+self.filename


    def set_config(self, **kwargs):

        """ Dynamically sets the configuration file """

        if "recdir" in kwargs:
            self.config.rec.recdir = kwargs["recdir"]
        if "subdirs" in kwargs:
            self.config.rec.subdirs = kwargs["subdirs"]
        if "label" in kwargs:
            self.config.rec.label = kwargs["label"]
        if "rectype" in kwargs:
            self.config.rec.rectype = kwargs["rectype"]

        if "rotation" in kwargs:
            self.config.cus.rotation = kwargs["rotation"]
        if "brighttune" in kwargs:
            self.config.cus.brighttune = kwargs["brighttune"]
        if "roi" in kwargs:
            self.config.cus.roi = kwargs["roi"]
        if "gains" in kwargs:
            self.config.cus.gains = kwargs["gains"]

        if "brightness" in kwargs:
            self.config.cam.brightness = kwargs["brightness"]
        if "contrast" in kwargs:
            self.config.cam.contrast = kwargs["contrast"]
        if "saturation" in kwargs:
            self.config.cam.saturation = kwargs["saturation"]
        if "iso" in kwargs:
            self.config.cam.iso = kwargs["iso"]
        if "sharpness" in kwargs:
            self.config.cam.sharpness = kwargs["sharpness"]
        if "compensation" in kwargs:
            self.config.cam.compensation = kwargs["compensation"]
        if "shutterspeed" in kwargs:
            self.config.cam.shutterspeed = kwargs["shutterspeed"]

        if "imgdims" in kwargs:
            self.config.img.imgdims = kwargs["imgdims"]
        if "viddims" in kwargs:
            self.config.vid.viddims = kwargs["viddims"]
        if "imgfps" in kwargs:
            self.config.img.imgfps = kwargs["imgfps"]
        if "vidfps" in kwargs:
            self.config.vid.vidfps = kwargs["vidfps"]

        if "imgwait" in kwargs:
            self.config.img.imgwait = kwargs["imgwait"]
        if "imgnr" in kwargs:
            self.config.img.imgnr = kwargs["imgnr"]
        if "imgtime" in kwargs:
            self.config.img.imgtime = kwargs["imgtime"]
        if "imgquality" in kwargs:
            self.config.img.imgquality = kwargs["imgquality"]

        if "vidduration" in kwargs:
            self.config.vid.vidduration = kwargs["vidduration"]
        if "viddelay" in kwargs:
            self.config.vid.viddelay = kwargs["viddelay"]
        if "vidquality" in kwargs:
            self.config.vid.vidquality = kwargs["vidquality"]

        brightchange = False
        if os.path.exists(self.brightfile):
            with open(self.brightfile) as f:
                brighttune = yaml.load(f, Loader=yaml.FullLoader)
                if brighttune != self.config.cus.brighttune:
                    self.config.cus.brighttune = brighttune
                    brightchange = True

        if len(kwargs) > 0 or brightchange:

            self._imgparams()
            self._shuttertofps()
            self.config.save()

            if "internal" not in kwargs:
                lineprint("Config settings stored and loaded..")


    def set_roi(self):

        """
        Dynamically draw a region of interest

        Explanation
        ===========
        This function will open a video stream of the raspberry pi camera. Enter
        'd' to start drawing the region of interest on an image taken from the
        video stream. When happy with the region selected, press 's' to store
        the coordinates, or 'esc' key to exit drawing on the image. To exist the
        video stream enter 'esc' key again.
        """

        C = Calibrate(internal=True, rotation=self.config.cus.rotation)
        if C.roi:
            self.set_config(roi=C.roi, internal="")
            lineprint("Roi stored..")
        else:
            lineprint("No roi selected..")


    def set_gains(self, auto = True):

        """Find the best gains for the raspberry pi camera"""

        if self.config.cus.roi == None:
            zoom = (0,0,1,1)
        else:
            zoom = literal_eval(self.config.cus.roi)
        (rg, bg) = setgains(startgains = checkfrac(self.config.cus.gains),
                            zoom = zoom, auto = auto)
        self.set_config(gains="(%5.2f, %5.2f)" % (rg, bg), internal="")
        lineprint("Gains: " + "(R:%5.2f, B:%5.2f)" % (rg, bg) + " stored..")


    def schedule(self, jobname = None, timeplan = None, enable = True,
                showjobs = False, clear = None, test = False):

        S = Schedule(jobname, timeplan, enable, showjobs, clear, test,
                     logfolder = self.logfolder, internal=True)


    def record(self):

        """Runs the Recorder instance"""

        self._setup_cam()
        self._namefile()

        if self.config.rec.rectype == "img":

            self.filename = self.filename + strftime("%H%M%S") + self.filetype
            self.cam.capture(self.filename, format="jpeg", resize = self.resize,
                             quality = self.config.img.imgquality)
            lineprint("Captured "+self.filename)

        elif self.config.rec.rectype == "imgseq":

            timepoint = datetime.now()
            for i, img in enumerate(self.cam.capture_continuous(self.filename,
                                    format="jpeg", resize = self.resize,
                                    quality = self.config.img.imgquality)):
                if i < self.config.img.imgnr-1:
                    timepassed = (datetime.now() - timepoint).total_seconds()
                    delay = max(0, self.config.img.imgwait - timepassed)
                    lineprint("Captured "+img+", sleeping "+str(round(delay,2))+"s..")
                    sleep(delay)
                    timepoint = datetime.now()
                else:
                    lineprint("Captured "+img)
                    break

        elif self.config.rec.rectype in ["vid","vidseq"]:

            # Temporary fix for flicker at start of (first) video..
            self.cam.start_recording(BytesIO(), format="h264", resize = self.resize)
            self.cam.wait_recording(2)
            self.cam.stop_recording()

            for session in ["_S%02d" % i for i in range(1,999)]:
                session = "" if self.config.rec.rectype == "vid" else session
                filename = self.filename+strftime("%H%M%S")+session+self.filetype
                self.cam.start_recording(filename, resize = self.resize,
                                         quality = self.config.vid.vidquality)
                lineprint("Start recording "+filename)
                self.cam.wait_recording(self.config.vid.vidduration + self.config.vid.viddelay)
                self.cam.stop_recording()
                lineprint("Finished recording "+filename)
                if self.config.rec.rectype == "vid":
                    break
                else:
                    if input("\nAny key for new session, e to exit: ") == "e":
                        break

        self.cam.close()


def rec():

    """To run pirecorder from the command line"""

    parser = argparse.ArgumentParser(prog="record",
    description="Runs PiRecorder record function")
    parser.add_argument("-c",
                        "--configfile",
                        default="pirecorder.conf",
                        action="store",
                        help="pirecorder configuration file")
    args = parser.parse_args()
    recorder = PiRecorder(args.configfile)
    recorder.set_config(internal=True)
    recorder.record()



#! /usr/bin/env python
"""
Copyright (c) 2019 - 2020 Jolle Jolles <j.w.jolles@gmail.com>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at:

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import cv2
import time
import argparse
import numpy as np

from pythutils.sysutils import lineprint
from pythutils.mediautils import checkroi, roi_to_zoom
import pythutils.drawutils as draw

from .videoin import VideoIn
from .__version__ import __version__

class Calibrate:

    def __init__(self, system="auto", framerate=8, vidsize=0.2, internal=False,
                 rotation = 0):

        """Opens a video stream with user interface to calibrate the camera"""

        if internal:
            lineprint("Running calibrate function.. ")

        self.system = system
        self.framerate = framerate
        self.vidsize = vidsize
        self.rotation = rotation
        self.cross = False
        self.stream = True
        self.exit = False
        self.roi = False
        self.fullscreen = False
        self.tempcol = draw.namedcols("orange")
        self.col = draw.namedcols("red")

        cv2.namedWindow("Image", cv2.WND_PROP_FULLSCREEN)
        self.m = draw.mouse_events()
        cv2.setMouseCallback('Image', self.m.draw)
        time.sleep(1)

        self.drawer()


    def draw_stream(self):

        lineprint("Streaming video..")

        self.vid = VideoIn(system=self.system, framerate=self.framerate,
                           vidsize=self.vidsize, rotation=self.rotation)
        self.vid.start()

        while True:
            self.img = self.vid.read()
            if self.cross:
                draw.draw_cross(self.img, pt2 = self.vid.res)
            cv2.imshow("Image", self.img)

            k = cv2.waitKey(1) & 0xFF
            if k == ord("c"):
                self.cross = not self.cross

            if k == ord("f"):
                self.fullscreen = not self.fullscreen
                if self.fullscreen:
                    cv2.setWindowProperty("Image",cv2.WND_PROP_FULLSCREEN,cv2.WINDOW_FULLSCREEN)
                else:
                    cv2.setWindowProperty("Image",cv2.WND_PROP_AUTOSIZE, cv2.WINDOW_NORMAL)
                    cv2.resizeWindow("Image", self.vid.res[0], self.vid.res[1])

            if k == ord("d"):
                self.stream = False
                break
            if k == 27:
                lineprint("User exited..")
                self.exit = True
                break

        self.vid.stop()


    def draw_frame(self):

        lineprint("Selecting roi..")
        self.imgbak = self.img.copy()

        while True:
            img = self.imgbak.copy()
            if self.m.twoPoint is not None:
                draw.draw_crosshair(img, self.m.pos)
            if self.m.posDown is not None:
                cv2.rectangle(img, self.m.posDown, self.m.pos, self.tempcol, 2)
            if self.m.posUp is not None:
                cv2.rectangle(img, self.m.pts[-1], self.m.posUp,
                              self.col, 2)
            cv2.imshow("Image", img)

            k = cv2.waitKey(1) & 0xFF
            if k == ord("s"):
                if self.m.twoPoint is not None:
                    self.m.twoPoint = checkroi(self.m.twoPoint, self.vid.res)
                    self.roi = roi_to_zoom(self.m.twoPoint, self.vid.res)
                    lineprint("roi "+str(self.roi)+" stored..")
                    break
                else:
                    lineprint("Nothing to store..")

            if k == ord("e"):
                self.roi = False
                lineprint("new roi erased..")

            if k == ord("z"):
                if self.m.twoPoint is not None:
                    lineprint("Creating zoomed image..")
                    self.vid2 = VideoIn(system=self.system, vidsize=self.vidsize,
                                        crop=self.m.twoPoint, rotation=self.rotation)
                    zimg = self.vid2.img()
                    cv2.namedWindow("Zoomed", cv2.WINDOW_NORMAL)
                    while True:
                        cv2.imshow("Zoomed", zimg)
                        cv2.resizeWindow("Zoomed", self.vid2.roiw, self.vid2.roih)
                        k = cv2.waitKey(1) & 0xFF
                        if k == 27:
                            break

                    k = 255
                    cv2.destroyWindow("Zoomed")

            if k == 27:
                self.stream = True
                self.m.twoPoint = None
                break


    def drawer(self):

        while True:
            if self.stream:
                self.draw_stream()
            if not self.stream:
                self.draw_frame()
            if self.exit:
                cv2.waitKey(1)
                cv2.destroyAllWindows()
                for i in range(5):
                    cv2.waitKey(1)
                break
